"""tap-yotpo abstract stream module."""
from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple

from singer import (
    Transformer,
    get_bookmark,
    get_logger,
    metrics,
    write_bookmark,
    write_record,
)
from singer.metadata import get_standard_metadata, to_list, to_map, write
from singer.utils import strftime, strptime_to_utc

LOGGER = get_logger()


class BaseStream(ABC):
    """
    A Base Class providing structure and boilerplate for generic streams
    and required attributes for any kind of stream
    ~~~
    Provides:
     - Basic Attributes (stream_name,replication_method,key_properties)
     - Helper methods for catalog generation
     - `sync` and `get_records` method for performing sync
    """

    @property
    @abstractmethod
    def stream(self) -> str:
        """Display Name of the stream."""

    @property
    @abstractmethod
    def tap_stream_id(self) -> str:
        """Unique identifier for the stream.

        This is allowed to be different from the name of the stream, in
        order to allow for sources that have duplicate stream names.
        """

    @property
    @abstractmethod
    def replication_method(self) -> str:
        """Defines the sync mode of a stream."""

    @property
    @abstractmethod
    def replication_key(self) -> str:
        """Defines the replication key for incremental sync mode of a
        stream."""

    @property
    @abstractmethod
    def valid_replication_keys(self) -> Tuple[str, str]:
        """Defines the replication key for incremental sync mode of a
        stream."""

    @property
    @abstractmethod
    def forced_replication_method(self) -> str:
        """Defines the sync mode of a stream."""

    @property
    @abstractmethod
    def key_properties(self) -> Tuple[str, str]:
        """List of key properties for stream."""

    @property
    def selected_by_default(self) -> bool:
        """Indicates if a node in the schema should be replicated, if a user
        has not expressed any opinion on whether or not to replicate it."""
        return False

    @property
    @abstractmethod
    def url_endpoint(self) -> str:
        """Defines the HTTP endpoint for the stream."""

    @abstractmethod
    def get_records(self):
        """Interacts with api client interaction and pagination."""

    @abstractmethod
    def sync(self, state: Dict, schema: Dict, stream_metadata: Dict, transformer: Transformer) -> Dict:
        """
        Performs a replication sync for the stream.
        ~~~
        Args:
         - state (dict): represents the state file for the tap.
         - schema (dict): Schema of the stream
         - transformer (object): A Object of the singer.transformer class.

        Returns:
         - bool: The return value. True for success, False otherwise.

        Docs:
         - https://github.com/singer-io/getting-started/blob/master/docs/SYNC_MODE.md#replication-method
        """

    def __init__(self, client=None) -> None:
        self.client = client

    @classmethod
    def get_metadata(cls, schema) -> Dict[str, str]:
        """Returns a `dict` for generating stream metadata."""
        stream_metadata = get_standard_metadata(
            **{
                "schema": schema,
                "key_properties": cls.key_properties,
                "valid_replication_keys": cls.valid_replication_keys,
                "replication_method": cls.replication_method or cls.forced_replication_method,
            }
        )
        stream_metadata = to_map(stream_metadata)
        if cls.valid_replication_keys is not None:
            for key in cls.valid_replication_keys:
                stream_metadata = write(stream_metadata, ("properties", key), "inclusion", "automatic")
        stream_metadata = to_list(stream_metadata)
        return stream_metadata


class IncrementalStream(BaseStream):
    """Base Class for Incremental Stream."""

    replication_method = "INCREMENTAL"
    forced_replication_method = "INCREMENTAL"
    config_start_key = None

    def get_bookmark(self, state: dict, key: Any = None) -> int:
        """A wrapper for singer.get_bookmark to deal with compatibility for
        bookmark values or start values."""
        return get_bookmark(
            state, self.tap_stream_id, key or self.replication_key, self.client.config.get(self.config_start_key, False)
        )

    def write_bookmark(self, state: dict, key: Any = None, value: Any = None) -> Dict:
        """A wrapper for singer.get_bookmark to deal with compatibility for
        bookmark values or start values."""
        return write_bookmark(state, self.tap_stream_id, key or self.replication_key, value)

    def sync(self, state: Dict, schema: Dict, stream_metadata: Dict, transformer: Transformer) -> Dict:
        """Abstract implementation for `type: Incremental` stream."""
        current_bookmark_date = self.get_bookmark(state)
        max_bookmark = current_bookmark_date_utc = strptime_to_utc(current_bookmark_date)

        with metrics.record_counter(self.tap_stream_id) as counter:
            for record in self.get_records():
                try:
                    record_timestamp = strptime_to_utc(record[self.replication_key])
                except IndexError as _:
                    LOGGER.error("Unable to process Record, Exception occurred: %s for stream %s", _, self.__class__)
                    continue
                if record_timestamp >= current_bookmark_date_utc:
                    transformed_record = transformer.transform(record, schema, stream_metadata)
                    write_record(self.tap_stream_id, transformed_record)
                    counter.increment()
                    max_bookmark = max(max_bookmark, record_timestamp)
                else:
                    LOGGER.warning("Skipping Record Older than the timestamp")

            state = self.write_bookmark(state, value=strftime(max_bookmark))
        return state


class FullTableStream(BaseStream):
    """Base Class for Incremental Stream."""

    replication_method = "FULL_TABLE"
    forced_replication_method = "FULL_TABLE"
    valid_replication_keys = None
    replication_key = None

    def sync(self, state: Dict, schema: Dict, stream_metadata: Dict, transformer: Transformer) -> Dict:
        """Abstract implementation for `type: Fulltable` stream."""
        with metrics.record_counter(self.tap_stream_id) as counter:
            for record in self.get_records():
                transformed_record = transformer.transform(record, schema, stream_metadata)
                write_record(self.tap_stream_id, transformed_record)
                counter.increment()
        return state


class UrlEndpointMixin:
    """A mixin for url formatting of URL's."""

    # pylint: disable=R0903; Mixin implementation

    url_endpoint: str = ""

    def get_url_endpoint(self) -> str:
        """Returns a formatted endpoint using the stream attributes."""
        return self.url_endpoint.replace("APP_KEY", self.client.config["api_key"])


class PageSizeMixin:
    """Adds a getter method to fetch page_size for current stream."""

    default_page_size = 100

    @property
    def page_size(self) -> int:
        """returns the `page_size` from config if present, else returns the
        self.default_page_size."""
        try:
            return int(getattr(self, "client").config.get("page_size", self.default_page_size))
        except (AttributeError):
            return self.default_page_size
